"""
Author: Pedro L. Magalhães
Email: pmlpm@posteo.de
Date: 2023
"""

#******************************************************************************
#******************************************************************************

# import libraries

import datetime as dt

import numpy as np

import math

import csv

#******************************************************************************
#******************************************************************************

# read csv file with tmy data

def read_jrc_tmy_csv(filename, assume_year: int = None):  
    """
    Reads a TMY file formatted according to the EC JRC's PVGIS (2022).
    
    This function reads a TMY CSV file formatted as those made available at the
    European Commission's JRC Photovoltaic Geographical Information System (or
    PVGIS) and returns the metadata and data in standard objects.    
    
    Parameters
    ----------
    
    filename : str
        A string identifying the file that is to be read.
        
    assume_year : int
        If this parameter is set to a non-default value, the dates will be set
        to the year indicated by this parameter ratheer than the original.
    
    Return
    ------
    
    data : numpy.ndarray
        A numpy ndarray containing the data read from the file. The data can be
        identified using the 'names' variable.
        
    names : list
        A list of strings identifying each of the fields in the array, namely:
            'timeUTC': for the datetime objects 
            'T2m': 2-m air temperature (degree Celsius)
            'RH': relative humidity (%)
            'Gh': Global irradiance on the horizontal plane (W/m2)
            'Gbn': Beam/direct irradiance on a plane always normal to sun rays (W/m2)
            'Gdh': Diffuse irradiance on the horizontal plane (W/m2)
            'IRh': Surface infrared (thermal) irradiance on a horizontal plane (W/m2)
            'WS10m': 10-m total wind speed (m/s)
            'WD10m': 10-m wind direction (0 = N, 90 = E) (degree)
            'SP': Surface (air) pressure (Pa)
                
    latitude : float
        The latitude (in radians north) of the place the file refers to.
        The latitude is between -pi/2 (south pole) and pi/2 (north pole).
        
    longitude : float
        The longitude (in radians west) of the place the file refers to.
        The longitude is between -2*pi and 2*pi and so multiple options exist.
        
    elevation : float
        The elevation (in meters) of the place the file refers to.
    
    Examples
    --------
    >>> filename = 'test_file.csv'
    >>> assume_year = 2012
    >>> data, names, lat, long, elev = read_jrc_tmy_csv(filename, assume_year)
        
    """
    
    #**************************************************************************
    
    # read the three first lines to extract the latitude, longitude and elev.
    
    with open(filename, newline='') as csvfile:
        
        reader = csv.reader(csvfile)
        
        # read the first three rows
        
        number_lines = 3
        
        rows = [reader.__next__()[0] for i in range(number_lines)]
        
        # find the location where the numbers start and convert them to floats
                
        number_prefix = ': '
        
        numbers = [
            np.float64(
                rows[i][rows[i].find(number_prefix)+len(number_prefix):])
            for i in range(number_lines)]
        
        # end
        
    # convert the latitude and the latitude
    
    latitude = numbers[0] # degrees north
    
    longitude = numbers[1] # degrees east
    
    elevation = numbers[2] # default: meters
    
    #**************************************************************************
    
    # file specifications
    
    number_header_lines = 16 # 17 but one line is for the names/labels
    
    number_footer_lines = 10 # 12 but two lines are empty
    
    file_encoding = "utf-8"
    
    date_format = '%Y%m%d:%H%M'
    
    tzinfo = dt.timezone.utc
    
    file_dtypes = (dt.datetime,
                   float,float,float,float,float,float,float,float,float)
    
    # handle the dates
    
    # differentiate date conversion depending on input parameters
    
    if assume_year != None and type(assume_year) == int:
        
        # set all entries to be for the same year, if requested
    
        date_converter = {
            0: lambda s: dt.datetime.strptime(
                s.decode(file_encoding),
                date_format).replace(year=assume_year,tzinfo=tzinfo)}
        
    else:
            
        # set the time zone only
    
        date_converter = {
            0: lambda s: dt.datetime.strptime(
                s.decode(file_encoding),
                date_format).replace(tzinfo=tzinfo)}
    
    #**************************************************************************
    
    # read the file
    
    npdata = np.genfromtxt(
        filename, 
        dtype=file_dtypes,
        names=True,
        skip_header=number_header_lines,
        skip_footer=number_footer_lines,
        delimiter=',',
        converters=date_converter)
            
    #**************************************************************************
    
    # get names
        
    names = ['timeUTC',
             'T2m',
             'RH',
             'Gh',
             'Gbn',
             'Gdh',
             'IRh',
             'WS10m',
             'WD10m',
             'SP']
    
    #**************************************************************************
    
    # return statement
    
    return ( 
        npdata, 
        names, 
        latitude, 
        longitude, 
        elevation
        )

    #**************************************************************************

# #******************************************************************************
# #******************************************************************************

# # get the standard longitude in degrees west from the local time

# def get_standard_meridian_longitude(lst_obj):
    
#     # lst_obj must have utc offset data
#     # a positive offset means east of greenwich
#     # a negative offset means west of greenwich
    
#     # every hour of difference is equivalent to 15º of longitude
#     # every second of difference is equivalent to 15ºC/3600 of longitude
#     # the total difference is time_in_seconds*15/3600
    
#     # if there is utc offset data
    
#     if lst_obj.utcoffset() != None:
        
#         # get the local standard meridian
    
#         # return math.radians(lst_obj.utcoffset().total_seconds()*15/3600)
    
#         return convertRealRadiansToPositiveRadians(
#             -math.radians(lst_obj.utcoffset().total_seconds()*15/3600)
#             )
    
#     else: # there is no time difference information
    
#         # print a warning
        
#         print('The local time object does not have timezone information.')
    
#         # return 0
    
#         return 0
    
#     #**************************************************************************
#     #**************************************************************************

# #******************************************************************************
# #******************************************************************************

# def convertRealRadiansToPositiveRadians(angle):
    
#     # angle is a real variable
#     # the output must be equivalent and between 0 and 2*pi
    
#     if angle >= 0: # positive angle
    
#         # longitude has to be between 0 and 2*pi
    
#         return math.fmod(angle,2*math.pi)
    
#     else: # negative angle
    
#         return math.fmod(angle,2*math.pi)+2*math.pi

#******************************************************************************
#******************************************************************************

def simplify_angle(angle: float,
                   angle_in_degrees: bool = True) -> float:
    
    if angle_in_degrees:
    
        return math.fmod(angle, 360)
    
    else:
        
        return math.fmod(angle, 2*math.pi)

#******************************************************************************
#******************************************************************************

def convert_angle(angle_degrees: float,
                  in_angle_positive_only: bool,
                  in_angle_positive_anticlockwise: bool,
                  out_angle_positive_only: bool,
                  out_angle_positive_anticlockwise: bool) -> float:
          
    #**************************************************************************
    
    # input format
    
    (in_format_one, 
     in_format_two, 
     in_format_three, 
     in_format_four) = _get_angle_format(in_angle_positive_only,
                                         in_angle_positive_anticlockwise)
                                          
    #**************************************************************************
    
    # check input angle conformity
    
    if angle_degrees < 0 and in_angle_positive_only:
        
        raise AngleFormatError 
        
    if abs(angle_degrees) > 180 and not in_angle_positive_only:

        raise AngleFormatError 
    
    #**************************************************************************
                                      
    # output format                                             
    
    (out_format_one, 
     out_format_two, 
     out_format_three, 
     out_format_four) = _get_angle_format(out_angle_positive_only,
                                          out_angle_positive_anticlockwise)

    #**************************************************************************
    
    # convert angle: 12 options
    
    # input format one
    
    if in_format_one and out_format_two:
        
        return 360-angle_degrees
    
    if in_format_one and out_format_three:
        
        return (
            angle_degrees 
            if angle_degrees <= 180 else angle_degrees-360)
    
    if in_format_one and out_format_four:
        
        return (
            -angle_degrees
            if angle_degrees < 180 else 360-angle_degrees)
    
    # input format two
    
    if in_format_two and out_format_one:
        
        return 360-angle_degrees
    
    if in_format_two and out_format_three:
        
        return (
            -angle_degrees
            if angle_degrees < 180 else 360-angle_degrees)
    
    if in_format_two and out_format_four:
        
        return (
            angle_degrees
            if angle_degrees <= 180 else angle_degrees-360)
    
    # input format three
    
    if in_format_three and out_format_one:
        
        return (
            angle_degrees
            if angle_degrees > 0 else 360+angle_degrees)
    
    if in_format_three and out_format_two:
        
        return (
            -angle_degrees
            if angle_degrees < 0 else 360-angle_degrees)
    
    if in_format_three and out_format_four:
        
        return -angle_degrees if angle_degrees != 180 else 180
    
    # input format four
    
    if in_format_four and out_format_one:
        
        return (
            -angle_degrees if angle_degrees < 0 else 360-angle_degrees)
    
    if in_format_four and out_format_two:
        
        return (
            angle_degrees if angle_degrees > 0 else 360+angle_degrees)
    
    if in_format_four and out_format_three:
        
        return -angle_degrees if angle_degrees != 180 else 180
    
    #**************************************************************************
    
#******************************************************************************
#******************************************************************************
    
def _get_angle_format(positive_only: bool,
                      positive_anticlockwise: bool) -> tuple:
    
    # format 1: positive only, positive anticlockwise
    # format 2: positive only, positive clockwise
    # format 3: positive or negative, positive anticlockwise
    # format 4: positive or negative, positive clockwise
    
    if positive_only:
        
        if positive_anticlockwise:
            
            format_one = True
            format_two = False
            format_three = False
            format_four = False
        
        else:
            
            format_one = False
            format_two = True
            format_three = False
            format_four = False
            
    else:
        
        if positive_anticlockwise:
            
            format_one = False
            format_two = False
            format_three = True
            format_four = False
        
        else:
            
            format_one = False
            format_two = False
            format_three = False
            format_four = True
            
    return format_one, format_two, format_three, format_four
            
#******************************************************************************
#******************************************************************************
            
class AngleFormatError(Exception):
    
    pass


"""
Author: Pedro L. Magalhães
Email: pmlpm@posteo.de
Date: 2023
"""

#******************************************************************************
#******************************************************************************

import datetime as dt

from pyproj import CRS, Transformer

# import math

#******************************************************************************
#******************************************************************************

class Location:
    """A class to specify a location on Earth."""
    
    crs = CRS.from_string('epsg:4326') 
    
    #crs = CRS.from_string('epsg:4326') 
    
    def __init__(self,
                 coordinates: tuple,
                 crs_string: str,
                 check_for_errors: bool = True
                 ):
        
        # check the number of coordinates
        
        if len(coordinates) != 2:
            
            raise ValueError('Incorrect number of coordinates.')
             
        # create internal CRS string
             
        self.crs_string = str.lower(self.crs.to_string())
        
        # verify type of CRS
        
        if (crs_string == self.crs_string or 
            crs_string == str.upper(self.crs_string)):
            
            # check the coordinates
            
            if not self.coordinates_are_within_bounds(*coordinates):
            
                raise ValueError('The coordinates are out of bounds.')
                
            # store them
            
            self.latitude, self.longitude = coordinates[0:2]
        
        else:
            
            # convert crs
            
            transformer = Transformer.from_crs(
                crs_from=crs_string, 
                crs_to=self.crs_string,
                always_xy=True)
            
            # convert coordinates
            
            self.latitude, self.longitude = transformer.transform(
                *coordinates,
                errcheck=check_for_errors)
            
    #**************************************************************************
            
    def coordinates_are_within_bounds(self, x: float, y: float):
        """
        Check if the coordinates are within the CRS\'s bounds.

        Parameters
        ----------
        x : float
            The first coordinate (e.g., the latitude with unprojected CRSs\').
        y : float
            The second coordinate (e.g., the longitude with unprojected CRSs\').

        Returns
        -------
        bool
            True if the coordinates within the bounds and False otherwise.

        """
        y_min, x_min, y_max, x_max = self.crs.area_of_use.bounds
        
        if x_min > x or x_max < x:
            
            return False
        
        if y_min > y or y_max < y:
            
            return False
        
        return True
    
    #**************************************************************************
    
    def local_standard_meridian_longitude(
            self, 
            local_time: dt.datetime, 
            dst_offset_seconds: float = 0.0) -> float:
        """
        Returns the longitude for the local standard meridian via local time.
        
        This longitude is determined via the time difference between the local
        time and UTC. The datetime object containing the time must be aware.

        Parameters
        ----------
        local_time : dt.datetime
            The datetime object containing the local time. It must be aware.
        dst_offset_seconds : float, optional
            The daylight savings time offset in seconds. The default is 0.0.
            This offset is only meant to provide a way to bypass the behaviour
            of datetime objects that do not have detailed daylight savings time
            information (e.g., objects merely with the correct UTC offset). Lo-
            cal times defined from an ISO string will tend to be incorrect for
            Summer months unless such an offset is used, since that is when an
            offset is added to the normal (i.e., Winter) UTC offset.

        Raises
        ------
        ValueError
            This exception is raised if the datetime object is not aware.

        Returns
        -------
        float
            The longitude for the local standard meridian in degrees.

        """
        
        # the object must be aware (e.g., to include the utcoffset method)
        
        if not is_datetime_aware(local_time):
            
            raise ValueError('The input datetime object is not aware.')

        # local_time must have utc offset data
        # a positive offset means east of greenwich
        # a negative offset means west of greenwich
        
        # every hour of difference is equivalent to 15º of longitude
        # every second of difference is equivalent to 15ºC/3600 of longitude
        # the total difference is time_in_seconds*15/3600
        
        return (
            (local_time.utcoffset().total_seconds()-dst_offset_seconds)*15/3600
            )

    #**************************************************************************
    #**************************************************************************

#******************************************************************************
#******************************************************************************

def day_of_the_year(my_time: dt.datetime,
                    as_integer: bool = True) -> int or float:
    """
    Returns the day of the year for a given datetime object.
    
    Since the actual day of the year differs, depending on the location, it is
    assumed that the day is determined according to UTC time.

    Parameters
    ----------
    my_time : dt.datetime
        The datetime object that defines the date/time under consideration.
    as_integer : bool, optional
        If True, returns the day of the year as an integer. If False, returns
        a float which also takes hours, minutes and seconds into account.
        The default is True.

    Returns
    -------
    int or float
        The day of the year corresponding to the date/time specified.

    """
    
    if as_integer:
    
        return my_time.utctimetuple().tm_yday
    
    else:
        
        return (
            my_time.utctimetuple().tm_yday+
            my_time.utctimetuple().tm_hour/24+
            my_time.utctimetuple().tm_min/(1440)+
            my_time.utctimetuple().tm_sec/(86400)
            )

#******************************************************************************
#******************************************************************************

def is_datetime_aware(my_time: dt.datetime) -> bool:
    """
    Verifies whether a datetime object is aware or not (naive).
    
    An aware object represents a specific moment in time that is not open to 
    interpretation. In turn, a naive object does not contain enough information
    to unambiguously locate itself relative to other date/time objects. 
    
    Thus, a datetime object d is aware if both of the following hold:
    1) d.tzinfo is not None
    2) d.tzinfo.utcoffset(d) does not return None
    Otherwise, d is naive.
    
    Source: https://docs.python.org/3/library/datetime.html

    Parameters
    ----------
    my_time : dt.datetime
        The datetime object under consideration.

    Returns
    -------
    bool
        A variable indicating the object is aware (True) or naive (False).

    """
    
    # A datetime object d is aware if both of the following hold:
    # d.tzinfo is not None
    # d.tzinfo.utcoffset(d) does not return None
    # Otherwise, d is naive.
    
    if my_time.tzinfo is not None:
        
        if my_time.tzinfo.utcoffset(my_time) != None:
            
            return True
        
    return False
    
#******************************************************************************
#******************************************************************************

class Plane:
    """A class for planes defined using azimuths and slopes."""
    
    def __init__(self,
                 azimuth_zero_south_positive_west: float,
                 slope_zero_horizontal_positive_south: float):
        
        self.azimuth = azimuth_zero_south_positive_west
        
        self.slope = slope_zero_horizontal_positive_south
    
#******************************************************************************
#******************************************************************************


"""
Author: Pedro L. Magalhães
Email: pmlpm@posteo.de
Date: 2023
"""

#******************************************************************************
#******************************************************************************

# import libraries

import math

# import pysoleng.solar as ps

from . import solar as ps

from .timeplace import Location, Plane, day_of_the_year

import datetime as dt

def tilted_plane_irradiance_isotropic(
        site: Location,
        surface: Plane,
        local_time: dt.datetime,
        site_ground_reflectance: float,
        site_horizontal_total,
        site_horizontal_diffuse,
        use_average_ratio: bool = True,
        sample_duration_seconds: int = 3600):    
    """
    Return the irradiance on tilted plane using the isotropic model.

    This function converts irradiation data obtained for a horizontal plane 
    into data that concerns a tilted plane, such as a wall or a solar module.    

    Parameters
    ----------
    site_longitude : float
        The site longitude in radians. If positive, it is west of Greenwich.
    site_latitude : float
        The site latitude in radians. If positive, it is north of the Equator.
    site_ground_reflectance : float
        The average ground reflectance (0<=,<=1) in the site\'s vinicity.
    plane_slope_horizontal : float
        The slope the target plane forms with the horizontal plane.
    plane_azimuth : float
        The azimuth the target plane has, in radians. If zero, it faces south.
        If pi/2, it faces west. If -pi/2, it faces east.
    star_altitude : float
        The local star\'s altitude as per the horizontal coordinate system.
        If zero, the star is on the horizon. If pi/2, it is overhead.
    star_azimuth : float
        The local star\' azimuth, as per the horizontal coordinate system.
        It uses the same convention as the plane_azimuth variable.
    site_time : datetime
        A datetime object for the local time, not the apparent solar time.
    site_horizontal_total : float
        The total irradiance on the horizontal plane for the site. It should
        concern the time period indicated by site_time.
    site_horizontal_diffuse : float
        The diffuse component of the total irradiance on the horizontal plane.
        It should concern the time period indicated by site_time.

    Return
    ------
    I_tilted_total : float
        Subset of the original series with the n first values.
    I_tilted_beam : float
        Subset of the original series with the n first values.
    I_tilted_sky_diffuse : float
        Subset of the original series with the n first values.
    I_tilted_albedo : float
        Subset of the original series with the n first values.

    See Also
    --------
    tail : Return the last n elements of the Series.

    Examples
    --------
    >>> s = pd.Series(['Ant', 'Bear', 'Cow', 'Dog', 'Falcon',
    ...                'Lion', 'Monkey', 'Rabbit', 'Zebra'])
    >>> s.head()
    0   Ant
    1   Bear
    2   Cow
    3   Dog
    4   Falcon
    dtype: object

    With the ``n`` parameter, we can change the number of returned rows:

    >>> s.head(n=3)
    0   Ant
    1   Bear
    2   Cow
    dtype: object
    """
    
    #**************************************************************************
    
    # verify inputs
    
    #**************************************************************************
    
    # convert the local time into solar time
    
    site_ast = ps.lst_to_ast(local_time, site)
    
    # get the declination angle
    
    declination_angle = math.radians(
        ps.declination_angle(
            day_of_the_year(local_time)
            )
        )
    
    # get the solar hour angle
    
    hour_angle = ps.hour_angle(site_ast)
    
    #**************************************************************************
    
    # compute ratio
    
    R_b = ps.beam_irradiation_ratio(
        declination_angle, 
        math.radians(site.latitude),   
        surface.slope, 
        surface.azimuth, 
        hour_angle,
        return_average_ratio=use_average_ratio,
        sample_duration_seconds=sample_duration_seconds) # half an hour each way
    
    #**************************************************************************
    
    # get the beam irradiation contribution to the global horiz. irradiation
    
    I_horizontal_beam = site_horizontal_total-site_horizontal_diffuse
    
    #**************************************************************************
    
    I_tilted_beam = R_b*I_horizontal_beam
    
    I_tilted_sky_diffuse = (
            site_horizontal_diffuse*
            (1+math.cos(surface.slope))/2
            )
    
    I_tilted_albedo = (
            site_horizontal_total*
            site_ground_reflectance*
            (1-math.cos(surface.slope))/2
            )
    
    I_tilted_total = I_tilted_beam+I_tilted_sky_diffuse+I_tilted_albedo

    #**************************************************************************
    #**************************************************************************

    # return function out
    
    return I_tilted_total, I_tilted_beam, I_tilted_sky_diffuse, I_tilted_albedo

#******************************************************************************
#******************************************************************************